"""
User Version Manager API
"""
from core_main_app.access_control.decorators import access_control
from core_main_app.components.template import api as template_api
from core_main_app.components.template.access_control import can_read_global
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.components.version_manager.utils import (
    get_latest_version_name,
)
from core_user_registration_app.components.user_template_version_manager.access_control import (
    can_write,
)
from core_user_registration_app.components.user_template_version_manager.models import (
    UserTemplateVersionManager,
)
import core_main_app.commons.exceptions as exceptions


@access_control(can_write)
def insert(user_version_manager, template, request):
    """Add a version to a user template version manager.

    Args:
        user_version_manager:
        template:
        request:

    Returns:

    """
    # save the template in database
    template_api.upsert(template, request=request)
    try:
        # create a version manager
        version_manager_api.upsert(user_version_manager, request=request)
        # set version manager
        template.version_manager = user_version_manager
        # set current version
        if len(user_version_manager.versions) == 0:
            template.is_current = True
        # update template
        template.display_name = get_latest_version_name(user_version_manager)
        # save template
        template.save_template()
        # return version manager
        return user_version_manager
    except exceptions.NotUniqueError as e:
        raise e
    except Exception as e:
        user_version_manager.delete()
        raise e


@access_control(can_read_global)
def get_global_version_managers(request, _cls=True):
    """Get all global version managers of a template.

    Args:
        request:
        _cls:

    Returns:

    """
    return UserTemplateVersionManager.get_global_version_managers(_cls)


def get_version_manager_by_id(id):
    """Return Version manager with given version manager id

    Args:
        id:

    Returns:

    """
    return UserTemplateVersionManager.get_by_id(id)


def get_default_version_manager():
    """Return default User Version Manager.

    Returns:

    """
    return UserTemplateVersionManager.objects.filter(is_default=True)


@access_control(can_write)
def upsert(user_version_manager, request):
    """Save or update user_version manager.

    Args:
        user_version_manager:
        request:

    Returns:

    """
    user_version_manager.save_version_manager()


def set_default_version_manager(user_version_manager, request):
    """Set default version manager.

    Args:
        user_version_manager:
        request:

    Returns:

    """
    default_user_version_manager = (
        UserTemplateVersionManager.get_default_version_manager()
    )
    for version_manager in default_user_version_manager:
        version_manager.is_default = False
        version_manager.save_version_manager()
    user_version_manager = get_version_manager_by_id(user_version_manager)
    user_version_manager.is_default = True
    user_version_manager.save_version_manager()

    return user_version_manager



"""Registration models
"""

from django.db import models

from core_main_app.permissions.utils import get_formatted_name
from core_user_registration_app.permissions import rights


class Register(models.Model):
    class Meta(object):
        verbose_name = "core_user_registration_app"
        default_permissions = ()
        permissions = (
            (rights.register_access, get_formatted_name(rights.register_access)),
            (
                rights.register_data_structure_access,
                get_formatted_name(rights.register_data_structure_access),
            ),
        )



"""
    Admin views
"""

from django.conf import settings
from django.contrib.admin.views.decorators import staff_member_required
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.html import escape as html_escape

import core_user_registration_app.components.account_request_metadata.api as account_request_api
from core_main_app.commons import exceptions
from core_main_app.components.template.models import Template
from core_main_app.components.template_version_manager import (
    api as template_version_manager_api,
)
from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.settings import INSTALLED_APPS
from core_main_app.utils.labels import get_data_label
from core_main_app.utils.rendering import admin_render
from core_main_app.utils.rendering import admin_render as render
from core_main_app.views.admin.forms import (
    UploadTemplateForm,
    UploadVersionForm,
)
from core_main_app.views.admin.views import handle_xsd_errors
from core_main_app.views.common.ajax import (
    EditTemplateVersionManagerView,
)
from core_main_app.views.common.views import CommonView
from core_main_app.views.common.views import read_xsd_file
from core_main_app.views.user.views import get_context_manage_template_versions
from core_user_registration_app.components.user_metadata import api as metadata_api
from core_user_registration_app.components.user_template_version_manager import (
    api as user_version_manager_api,
)
from core_user_registration_app.components.user_template_version_manager.models import (
    UserTemplateVersionManager,
)


@staff_member_required
def user_requests(request):
    """Page that allows to accept or deny user requests

    Args:
        request:

    Returns:
    """
    # Call the API
    user = request.user
    requests = account_request_api.get_all(user)

    assets = {
        "js": [
            {"path": "core_website_app/admin/js/user_requests.js", "is_raw": False},
            {
                "path": "core_user_registration_app/admin/js/view_meta.js",
                "is_raw": False,
            },
            {
                "path": "core_user_registration_app/admin/js/view_meta.raw.js",
                "is_raw": True,
            },
        ],
    }

    modals = [
        "core_website_app/admin/account_requests/modals/deny_request.html",
    ]

    return render(
        request,
        "core_user_registration_app/admin/user_requests.html",
        assets=assets,
        modals=modals,
        context={"requests": requests},
    )


class ViewMetaData(CommonView):
    """
    View detail data.
    """

    template = "core_main_app/user/data/detail.html"

    @method_decorator(staff_member_required)
    def get(self, request, *args, **kwargs):
        data_id = request.GET["id"]
        try:
            data = metadata_api.get_by_id(data_id, request.user)

            context = {"data": data}

            assets = {
                "js": [
                    {"path": "core_main_app/common/js/XMLTree.js", "is_raw": False},
                    {"path": "core_main_app/user/js/data/detail.js", "is_raw": False},
                ],
                "css": ["core_main_app/common/css/XMLTree.css"],
            }

            modals = []

            if "core_file_preview_app" in INSTALLED_APPS:
                assets["js"].extend(
                    [
                        {
                            "path": "core_file_preview_app/user/js/file_preview.js",
                            "is_raw": False,
                        }
                    ]
                )
                assets["css"].append("core_file_preview_app/user/css/file_preview.css")
                modals.append("core_file_preview_app/user/file_preview_modal.html")

            return self.common_render(
                request, self.template, context=context, assets=assets, modals=modals
            )
        except exceptions.DoesNotExist:
            error_message = "Data not found"
        except exceptions.ModelError:
            error_message = "Model error"
        except Exception as e:
            error_message = str(e)

        return self.common_render(
            request,
            "core_main_app/common/commons/error.html",
            context={
                "error": "Unable to access the requested "
                + get_data_label()
                + ": {}.".format(error_message)
            },
        )


@staff_member_required
def manage_user_templates(request):
    """View that allows user template management.

    Args:
        request:

    Returns:

    """
    templates = user_version_manager_api.get_global_version_managers(request=request)
    context = {
        "object_name": "Template",
        "available": [template for template in templates if not template.is_disabled],
        "disabled": [template for template in templates if template.is_disabled],
        "come_from_user": "True",
    }

    assets = {
        "js": [
            {
                "path": "core_main_app/common/js/templates/list/restore_user.js",
                "is_raw": False,
            },
            {
                "path": "core_main_app/common/js/templates/list/modals/disable_user.js",
                "is_raw": False,
            },
            {
                "path": "core_user_registration_app/common/js/templates/set_current.js",
                "is_raw": False,
            },
            EditTemplateVersionManagerView.get_modal_js_path(),
        ]
    }

    modals = [
        "core_main_app/admin/templates/list/modals/disable.html",
        EditTemplateVersionManagerView.get_modal_html_path(),
    ]

    return admin_render(
        request,
        "core_user_registration_app/admin/templates/list.html",
        assets=assets,
        context=context,
        modals=modals,
    )


@staff_member_required
def manage_user_template_versions(request, version_manager_id):
    """View that allows template versions management.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    try:
        # get the version manager

        version_manager = user_version_manager_api.get_version_manager_by_id(
            version_manager_id
        )
        context = get_context_manage_template_versions(version_manager)
        if "core_parser_app" in settings.INSTALLED_APPS:
            context.update({"module_url": "admin:core_parser_app_template_modules"})

        assets = {
            "js": [
                {
                    "path": "core_user_registration_app/common/js/templates/versions/set_current.js",
                    "is_raw": False,
                },
                {
                    "path": "core_user_registration_app/common/js/templates/versions/set_current.raw.js",
                    "is_raw": True,
                },
                {
                    "path": "core_main_app/common/js/templates/versions/restore.js",
                    "is_raw": False,
                },
                {
                    "path": "core_main_app/common/js/templates/versions/modals/disable.js",
                    "is_raw": False,
                },
            ]
        }

        modals = ["core_main_app/admin/templates/versions/modals/disable.html"]

        return admin_render(
            request,
            "core_user_registration_app/admin/templates/versions.html",
            assets=assets,
            modals=modals,
            context=context,
        )
    except Exception as e:
        return admin_render(
            request,
            "core_main_app/common/commons/error.html",
            context={"error": str(e)},
        )


@staff_member_required
def upload_template(request):
    """Upload template.

    Args:
        request:

    Returns:

    """
    assets = {
        "js": [
            {
                "path": "core_main_app/admin/js/templates/upload/dependency_resolver.js",
                "is_raw": True,
            },
            {
                "path": "core_main_app/admin/js/templates/upload/dependencies.js",
                "is_raw": False,
            },
            {"path": "core_main_app/common/js/backtoprevious.js", "is_raw": True},
        ]
    }

    context = {
        "object_name": "Template",
        "url": reverse("core-admin:core_user_registration_app_upload_template"),
        "redirect_url": reverse("core-admin:core_user_registration_app_templates"),
    }

    # method is POST
    if request.method == "POST":
        form = UploadTemplateForm(request.POST, request.FILES)
        context["upload_form"] = form

        if form.is_valid():
            return _save_template(request, assets, context)
        else:
            # Display error from the form
            return _upload_template_response(request, assets, context)
    # method is GET
    else:
        # render the form to upload a template
        context["upload_form"] = UploadTemplateForm()
        return _upload_template_response(request, assets, context)


@staff_member_required
def upload_template_version(request, version_manager_id):
    """Upload template version.

    Args:
        request:
        version_manager_id:

    Returns:

    """
    assets = {
        "js": [
            {
                "path": "core_main_app/admin/js/templates/upload/dependency_resolver.js",
                "is_raw": True,
            },
            {
                "path": "core_main_app/admin/js/templates/upload/dependencies.js",
                "is_raw": False,
            },
        ]
    }

    template_version_manager = version_manager_api.get(
        version_manager_id, request=request
    )
    context = {
        "object_name": "Template",
        "version_manager": template_version_manager,
        "url": reverse(
            "admin:core_main_app_upload_template_version",
            kwargs={"version_manager_id": template_version_manager.id},
        ),
        "redirect_url": reverse(
            "admin:core_main_app_manage_template_versions",
            kwargs={"version_manager_id": template_version_manager.id},
        ),
    }

    # method is POST
    if request.method == "POST":
        form = UploadVersionForm(request.POST, request.FILES)
        context["upload_form"] = form

        if form.is_valid():
            return _save_template_version(
                request, assets, context, template_version_manager
            )
        else:
            # Display errors from the form
            return _upload_template_response(request, assets, context)
    # method is GET
    else:
        # render the form to upload a template
        context["upload_form"] = UploadVersionForm()
        return _upload_template_response(request, assets, context)


@staff_member_required
def _save_template(request, assets, context):
    """Save a template.

    Args:
        request:
        assets:
        context:

    Returns:

    """
    # get the schema name
    name = request.POST["name"]
    # get the file from the form
    xsd_file = request.FILES["upload_file"]
    # read the content of the file
    xsd_data = read_xsd_file(xsd_file)
    try:
        template = Template(filename=xsd_file.name, content=xsd_data)
        template_version_manager = UserTemplateVersionManager(title=name)
        user_version_manager_api.insert(
            template_version_manager, template, request=request
        )
        return HttpResponseRedirect(
            reverse("core-admin:core_user_registration_app_templates")
        )
    except exceptions.XSDError as xsd_error:
        return handle_xsd_errors(
            request, assets, context, xsd_error, xsd_data, xsd_file.name
        )
    except exceptions.NotUniqueError:
        context["errors"] = html_escape(
            "A template with the same name already exists. Please choose another name."
        )
        return _upload_template_response(request, assets, context)
    except Exception as e:
        context["errors"] = html_escape(str(e))
        return _upload_template_response(request, assets, context)


@staff_member_required
def _save_template_version(request, assets, context, template_version_manager):
    """Save a template version.

    Args:
        request:
        assets:
        context:
        template_version_manager:

    Returns:

    """
    # get the file from the form
    xsd_file = request.FILES["xsd_file"]
    # read the content of the file
    xsd_data = read_xsd_file(xsd_file)

    try:
        template = Template(filename=xsd_file.name, content=xsd_data)
        template_version_manager_api.insert(
            template_version_manager, template, request=request
        )

        # create the fragment url with all the version of the template (minus the new template)
        version_manager_string = ""
        for version in template_version_manager.versions:
            if version != str(template.id):
                current_version_string = (
                    version if version_manager_string == "" else f",{version}"
                )

                version_manager_string += current_version_string

        # add the fragment data to the url
        fragment = f"#from={version_manager_string}&to={template.id}"

        return HttpResponseRedirect(
            reverse("admin:core_main_app_data_migration") + fragment
        )
    except exceptions.XSDError as xsd_error:
        return handle_xsd_errors(
            request, assets, context, xsd_error, xsd_data, xsd_file.name
        )
    except Exception as e:
        context["errors"] = html_escape(str(e))
        return _upload_template_response(request, assets, context)


@staff_member_required
def _upload_template_response(request, assets, context):
    """Render template upload response.

    Args:
        request:
        context:

    Returns:

    """
    return admin_render(
        request,
        "core_main_app/admin/templates/upload.html",
        assets=assets,
        context=context,
    )



""" Unit Test UserMetadata
"""

import datetime

import pytz

from core_main_app.commons import exceptions
from core_main_app.settings import DATA_SORTING_FIELDS
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_user_registration_app.components.user_metadata.api import (
    check_xml_file_is_valid,
)
from core_user_registration_app.components.user_metadata.models import UserMetadata
from tests.components.metadata.fixtures.fixtures import (
    DataFixtures,
    AccessControlDataFixture,
)

fixture_data = DataFixtures()
access_control_data_fixture = AccessControlDataFixture()


class TestDataGetAll(MongoIntegrationBaseTestCase):

    fixture = access_control_data_fixture

    def test_data_get_all_return_collection_of_data(self):
        # Act
        result = UserMetadata.get_all(DATA_SORTING_FIELDS)
        # Assert
        self.assertTrue(all(isinstance(item, UserMetadata) for item in result))

    def test_data_get_all_return_objects_data_in_collection(self):
        # Act
        result = UserMetadata.get_all(DATA_SORTING_FIELDS)
        # Assert
        self.assertTrue(len(self.fixture.data_collection) == result.count())

    def test_data_get_all_ordering(self):
        # Arrange
        access_control_data_fixture.insert_data()
        ascending_order_by_field = ["+title"]
        descending_order_by_field = ["-title"]
        # Act
        ascending_result = UserMetadata.get_all(ascending_order_by_field)
        descending_result = UserMetadata.get_all(descending_order_by_field)
        # Assert
        for i in range(len(ascending_result)):
            self.assertTrue(
                ascending_result.all()[i].title
                == descending_result.all()[len(ascending_result) - i - 1].title
            )

    def test_data_get_all_ascending_sorting(self):
        # Arrange
        ascending_order_by_field = ["+title"]
        # Act
        ascending_result = UserMetadata.get_all(ascending_order_by_field)
        # Assert
        self.assertTrue(self.fixture.data_1.title == ascending_result.all()[0].title)
        self.assertTrue(self.fixture.data_2.title == ascending_result.all()[1].title)

    def test_data_get_all_descending_sorting(self):
        # Arrange
        descending_order_by_field = ["-title"]
        # Act
        descending_result = UserMetadata.get_all(descending_order_by_field)
        # Assert
        self.assertTrue(
            self.fixture.data_2.title
            == descending_result.all()[len(descending_result) - 2].title
        )
        self.assertTrue(
            self.fixture.data_1.title
            == descending_result.all()[len(descending_result) - 1].title
        )

    def test_data_get_all_multi_field_sorting(self):
        # Arrange
        ascending_order_by_multi_field = ["+title", "+user_id"]
        descending_order_by_multi_field = ["+title", "-user_id"]
        # Act
        ascending_result = UserMetadata.get_all(ascending_order_by_multi_field)
        descending_result = UserMetadata.get_all(descending_order_by_multi_field)
        # Assert
        self.assertEqual(self.fixture.data_4.user_id, ascending_result.all()[4].user_id)
        self.assertEqual(self.fixture.data_5.user_id, ascending_result.all()[3].user_id)

        self.assertEqual(
            self.fixture.data_4.user_id, descending_result.all()[3].user_id
        )
        self.assertEqual(
            self.fixture.data_5.user_id, descending_result.all()[4].user_id
        )


class TestDataGetById(MongoIntegrationBaseTestCase):

    fixture = fixture_data

    def test_data_get_by_id_raises_api_error_if_not_found(self):
        # Act # Assert
        with self.assertRaises(exceptions.DoesNotExist):
            UserMetadata.get_by_id(-1)

    def test_data_get_by_id_return_data_if_found(self):
        # Act
        result = UserMetadata.get_by_id(self.fixture.data_1.id)
        # Assert
        self.assertEqual(result, self.fixture.data_1)


def mock_upsert(UserMetadata, user):
    if UserMetadata.xml_content is None:
        raise exceptions.ApiError(
            "Unable to save UserMetadata: xml_content field is not set."
        )

    UserMetadata.last_modification_date = datetime.datetime.now(pytz.utc)
    check_xml_file_is_valid(UserMetadata)
    return UserMetadata.save()


def _create_user(user_id, is_superuser=False):
    return create_mock_user(user_id, is_superuser=is_superuser)



""" User data Structure api
"""
from core_main_app.access_control.api import has_perm_administration
from core_main_app.access_control.decorators import access_control
from core_user_registration_app.access_control.api import (
    can_read,
    can_change_owner,
)
from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)


def upsert(user_data_structure, user):
    """Save or update the user Data Structure

    Args:
        user_data_structure:
        user:

    Returns:

    """
    return user_data_structure.save_object()


@access_control(has_perm_administration)
def get_all(user):
    """Returns all user data structure api

    Returns:

    """
    return UserDataStructure.get_all()


def get_by_id(user_data_structure_id):
    """Returns the user data structure with the given id

    Args:
        user_data_structure_id:
        user:

    Returns:

    """
    return UserDataStructure.get_by_id(user_data_structure_id)


def delete(user_data_structure, user):
    """Deletes the user data structure and the element associated

    Args:
        user_data_structure:
        user:
    """
    user_data_structure.delete()


@access_control(can_read)
def get_by_data_id(data_id, user):
    """Return the user data structure with the given data id

    Args:
        data_id:

    Returns:

    """
    return UserDataStructure.get_by_data_id(data_id)


def update_data_structure_root(user_data_structure, root_element, user):
    """Update the data structure with a root element.

    Args:
        user_data_structure:
        root_element:

    Returns:

    """
    # Delete data structure elements
    user_data_structure.delete_data_structure_elements_from_root()

    # set the root element in the data structure
    user_data_structure.data_structure_element_root = root_element

    # save the data structure
    return upsert(user_data_structure, user)


@access_control(can_read)
def get_by_data_structure_element_root_id(data_structure_element_root, user):
    """Return the user data structure with the given data structure element root id

    Args:
        data_structure_element_root:
        user:

    Returns:

    """
    return UserDataStructure.get_by_data_structure_element_root(
        data_structure_element_root
    )


@access_control(can_change_owner)
def change_owner(user_data_structure, new_user, user):
    """Change user data structure's owner.

    Args:
        user_data_structure:
        user:
        new_user:

    Returns:
    """
    user_data_structure.user = str(new_user.id)
    user_data_structure.save_object()



"""User Data Structure models
"""
from django.core.exceptions import ObjectDoesNotExist
from django.db.models.signals import pre_delete


from core_main_app.commons import exceptions
from core_main_app.components.data.models import Data
from core_parser_app.components.data_structure.models import DataStructure
from core_user_registration_app.permissions import rights
from django.db import models, IntegrityError


class UserDataStructure(DataStructure):
    """user data structure."""

    form_string = models.TextField(blank=True)
    data = models.ForeignKey(Data, blank=True, on_delete=models.CASCADE, null=True)

    @staticmethod
    def get_permission():
        return f"{rights.register_content_type}.{rights.register_data_structure_access}"

    def save_object(self):
        """Custom save

        Returns:

        """
        try:
            return self.save()
        except IntegrityError:
            raise exceptions.ModelError("Unable to save the document: not unique.")
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def get_by_id(data_structure_id):
        """Return the object with the given id.

        Args:
            data_structure_id:

        Returns:
            user Data Structure (obj): userDataStructure object with the given id

        """
        try:
            return UserDataStructure.objects.get(pk=str(data_structure_id))
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def get_none():
        """Return None object, used by forms.

        Returns:

        """
        return UserDataStructure.objects.none()

    @staticmethod
    def get_all():
        """Return all user data structure api.

        Returns:

        """
        return UserDataStructure.objects.all()

    @staticmethod
    def get_all_by_user_id_and_template_id(user_id, template_id):
        """Return all template version managers with user set to None.

        Returns:

        """
        return UserDataStructure.objects.filter(
            user=str(user_id), template=str(template_id)
        ).all()

    @staticmethod
    def get_by_user_id_and_template_id_and_name(user_id, template_id, name):
        """Return the user data structure with user, template id and name.

        Returns:

        """
        try:
            return UserDataStructure.objects.get(
                user=str(user_id), template=str(template_id), name=name
            )
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def get_all_by_user_id_with_no_data(user_id):
        """
        Return all the user date structure of the user, with no data.

        Args: user_id:
        Return:
        """
        return UserDataStructure.objects.filter(
            user=str(user_id), data__isnull=True
        ).all()

    @staticmethod
    def get_all_except_user_id_with_no_data(user_id):
        """Return all the user date structure except the one of the user, with no data.

        Args: user_id:
        Return:
        """
        return (
            UserDataStructure.objects.exclude(user=str(user_id))
            .filter(data__isnull=True)
            .all()
        )

    @staticmethod
    def get_all_by_user_id_and_template_id_with_no_data(user_id, template_id):
        """

        Args:
            user_id:
            template_id:

        Returns:

        """
        return UserDataStructure.objects.filter(
            user=str(user_id), template=str(template_id), data__isnull=True
        ).all()

    @staticmethod
    def get_all_with_no_data():
        """Returns all user data structure api with no link to a data.

        Args:

        Returns:

        """
        return UserDataStructure.objects.filter(data__isnull=True).all()

    @staticmethod
    def get_all_by_user(user_id):
        """Return all user data structure by user.

        Returns:

        """
        return UserDataStructure.objects.filter(user=str(user_id)).all()

    @staticmethod
    def get_by_data_id(data_id):
        """Return the user data structure with the given data id

        Args:
            data_id:

        Returns:

        """
        try:
            return UserDataStructure.objects.get(data=str(data_id))
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    @staticmethod
    def get_by_data_structure_element_root(data_structure_element_root):
        """Return the user data structure with the given data id

        Args:
            data_structure_element_root:

        Returns:

        """
        try:
            return UserDataStructure.objects.get(
                data_structure_element_root=str(data_structure_element_root.id)
            )
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))


# Connect signals
pre_delete.connect(DataStructure.pre_delete, sender=UserDataStructure)



""" Set of functions to define the common rules for access control across collections
"""
import logging

from django.contrib.auth.models import User, AnonymousUser

from core_main_app.access_control.exceptions import AccessControlError
from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)

logger = logging.getLogger(__name__)


def can_read(func, *args, **kwargs):
    """Can user read

    Args:
        func:
        *args:
        **kwargs:

    Returns:

    """
    # get the User in the args list
    user = next(
        (
            arg
            for arg in args
            if isinstance(arg, User) or isinstance(arg, AnonymousUser)
        ),
        None,
    )

    if user.is_anonymous or user is None:
        raise AccessControlError(
            "The user doesn't have enough rights to access this document."
        )

    if user.is_superuser:
        return func(*args, **kwargs)

    document = func(*args, **kwargs)
    if type(document) is list:
        document_list = document
    elif isinstance(document, UserDataStructure):
        document_list = [document]
    else:
        document_list = list(document)

    _check_can_read(document_list, user)

    return document


def _check_can_read(document_list, user):
    """Can read from list id.

    Args:
        document_list:
        user:

    Returns:

    """
    # check access is correct
    for document in document_list:
        # user is document owner
        if document.user == str(user.id):
            continue
        # user is not owner or document
        raise AccessControlError("The user doesn't have enough rights.")


def can_write(func, *args, **kwargs):
    """Can user write

    Args:
        func:
        *args:
        **kwargs:

    Returns:

    """
    user = next(
        (
            arg
            for arg in args
            if isinstance(arg, User) or isinstance(arg, AnonymousUser)
        ),
        None,
    )

    if user.is_anonymous or user is None:
        raise AccessControlError(
            "The user doesn't have enough rights to access this document."
        )

    if user.is_superuser:
        return func(*args, **kwargs)

    document = next((arg for arg in args if isinstance(arg, UserDataStructure)), None)
    if document.user != str(user.id):
        raise AccessControlError(
            "The user doesn't have enough rights to access this document."
        )

    return func(*args, **kwargs)


def can_change_owner(func, document, new_user, user):
    """Can user change document's owner.

    Args:
        func:
        document:
        new_user:
        user:

    Returns:

    """
    if user.is_anonymous:
        raise AccessControlError(
            "The user doesn't have enough rights to access this document."
        )
    if user.is_superuser:
        return func(document, new_user, user)

    if document.user == str(user.id):
        return func(document, new_user, user)

    raise AccessControlError(
        "The user doesn't have enough rights to access this document."
    )



# Generated by Django 3.2.10 on 2022-03-18 13:53

import django.contrib.postgres.search
import django.core.validators
import django.db.models.deletion
from django.db import migrations, models

import core_main_app


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("core_website_app", "0001_initial"),
        ("core_parser_app", "0001_initial"),
        ("core_main_app", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Register",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "core_user_registration_app",
                "permissions": (
                    ("access_register", "Can access register"),
                    ("access_user_data_structure", "Can access user data structure"),
                ),
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="UserTemplateVersionManager",
            fields=[
                (
                    "templateversionmanager_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core_main_app.templateversionmanager",
                    ),
                ),
                ("is_default", models.BooleanField(default=False)),
            ],
            options={
                "abstract": False,
            },
            bases=("core_main_app.templateversionmanager",),
        ),
        migrations.CreateModel(
            name="UserMetadata",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("dict_content", models.JSONField(blank=True, null=True)),
                (
                    "title",
                    models.CharField(
                        max_length=200,
                        validators=[
                            django.core.validators.RegexValidator(
                                code="invalid_title",
                                message="Title must not be empty or only whitespaces",
                                regex=".*\\S.*",
                            )
                        ],
                    ),
                ),
                (
                    "xml_file",
                    models.FileField(
                        upload_to=core_main_app.utils.storage.storage.user_directory_path
                    ),
                ),
                (
                    "vector_column",
                    django.contrib.postgres.search.SearchVectorField(null=True),
                ),
                (
                    "creation_date",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "last_modification_date",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "last_change_date",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                ("user_id", models.CharField(max_length=200)),
                (
                    "template",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core_main_app.template",
                    ),
                ),
                (
                    "workspace",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="core_main_app.workspace",
                    ),
                ),
            ],
            options={
                "verbose_name": "User metadata",
                "verbose_name_plural": "User metadata",
            },
        ),
        migrations.CreateModel(
            name="UserDataStructure",
            fields=[
                (
                    "datastructure_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core_parser_app.datastructure",
                    ),
                ),
                ("form_string", models.TextField(blank=True)),
                (
                    "data",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core_main_app.data",
                    ),
                ),
            ],
            bases=("core_parser_app.datastructure",),
        ),
        migrations.CreateModel(
            name="AccountRequestMetadata",
            fields=[
                (
                    "accountrequest_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core_website_app.accountrequest",
                    ),
                ),
                (
                    "metadata",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core_user_registration_app.usermetadata",
                    ),
                ),
            ],
            options={
                "verbose_name": "Account request metadata",
                "verbose_name_plural": "Account request metadata",
            },
            bases=("core_website_app.accountrequest",),
        ),
        migrations.AddIndex(
            model_name="usermetadata",
            index=models.Index(
                fields=["title", "last_modification_date", "template", "user_id"],
                name="core_user_r_title_cc1330_idx",
            ),
        ),
    ]



""" AccountRequestMetadata model
"""

from django.db import models

from core_user_registration_app.components.user_metadata.models import UserMetadata
from core_website_app.components.account_request.models import AccountRequest


class AccountRequestMetadata(AccountRequest):
    """AccountRequestMetadata."""

    metadata = models.ForeignKey(
        UserMetadata, blank=True, on_delete=models.CASCADE, null=True
    )

    class Meta:
        verbose_name = "Account request metadata"
        verbose_name_plural = "Account request metadata"

    @staticmethod
    def get_all():
        """Get all Account Request Metadata

        Returns:

        """
        return AccountRequestMetadata.objects.all()



""" REST views for the registry template version manager API
"""
from django.contrib.admin.views.decorators import staff_member_required
from django.http.response import HttpResponseBadRequest, HttpResponse
from rest_framework.permissions import IsAuthenticated

from core_main_app.components.version_manager import api as version_manager_api
from core_main_app.rest.template_version_manager.abstract_views import (
    AbstractStatusTemplateVersion,
)
from core_user_registration_app.components.user_template_version_manager import (
    api as user_version_manager_api,
)


# ACL is done in the AbstractStatusTemplateVersion object
class CurrentTemplateVersion(AbstractStatusTemplateVersion):
    """Update status to current"""

    permission_classes = (IsAuthenticated,)

    def status_update(self, template_object):
        """Update status to current

        Args:

            template_object: template_version

        Returns:

            TemplateVersion
        """
        return version_manager_api.set_current(template_object, request=self.request)


@staff_member_required
def set_current_template_version_from_version_manager(request):
    """Set the current version of a template.

    Args:
        request:

    Returns:

    """
    try:

        user_version_manager_api.set_default_version_manager(request.GET["id"], request)
    except Exception as e:
        return HttpResponseBadRequest(str(e), content_type="application/javascript")

    return HttpResponse(content_type="application/javascript")



""" Initialize permissions for core user registration app.
"""
import logging

from django.contrib.auth.models import Group, Permission
from django.core.exceptions import ObjectDoesNotExist
from django_celery_beat.models import CrontabSchedule, PeriodicTask

import core_main_app.permissions.rights as main_rights
import core_user_registration_app.permissions.rights as registration_rights
from core_user_registration_app.components.user_template_version_manager import (
    api as user_version_manager_api,
)
from core_user_registration_app.settings import (
    REGISTRY_XSD_USER_FILENAME,
    REGISTRY_XSD_USER_FILEPATH,
)
from core_user_registration_app.system import api as registry_system_api
from core_user_registration_app.tasks import delete_user_data_structure

logger = logging.getLogger(__name__)

from os.path import join

from django.contrib.staticfiles import finders

from core_user_registration_app.settings import (
    XSL_FOLDER_PATH,
    LIST_XSL_FILENAME,
    DETAIL_XSL_FILENAME,
)
from core_main_app.commons import exceptions
from core_main_app.system import api as system_api
from core_main_app.components.xsl_transformation import api as xslt_transformation_api
from core_main_app.components.xsl_transformation.models import XslTransformation
from core_main_app.utils.file import read_file_content


def init_registration_app():
    """Init the registry. Add the registry template.

    Returns:

    """

    try:
        # Init scheduled tasks
        _init_periodic_tasks()
        # Add template
        _add_user_template()
        # Init the permissions
        _init_permissions()
        # Init the xslt
        _init_xslt()
    except Exception as e:
        logger.error("Impossible to init the registration app: {0}".format(str(e)))


def _init_permissions():
    """Initialization of groups and permissions."""
    try:
        # Get or Create the default group
        anonymous, created = Group.objects.get_or_create(
            name=main_rights.anonymous_group
        )

        # Get registration permissions
        register_access_perm = Permission.objects.get(
            codename=registration_rights.register_access
        )
        register_view_data_save_repo_perm = Permission.objects.get(
            codename=registration_rights.register_data_structure_access
        )

        # Add permissions to default group
        anonymous.permissions.add(
            register_access_perm,
            register_view_data_save_repo_perm,
        )

    except Exception as e:
        logger.error("Impossible to init register permissions: %s" % str(e))


def _add_user_template():
    """Add the registry template.

    Returns:

    """
    xsd_filepath = REGISTRY_XSD_USER_FILEPATH
    xsd_filename = REGISTRY_XSD_USER_FILENAME
    if xsd_filename == "":
        raise Exception(
            "Please configure the REGISTRY_XSD_USER_FILENAME setting in your project."
        )
    if xsd_filepath == "":
        raise Exception(
            "Please configure the REGISTRY_XSD_USER_FILEPATH setting in your project."
        )
    try:
        registry_system_api.get_active_global_version_manager_by_title(xsd_filename)
    except exceptions.DoesNotExist:
        default_xsd_path = finders.find(xsd_filepath)
        xsd_data = read_file_content(default_xsd_path)
        registry_system_api.insert_registry_user_schema(xsd_filename, xsd_data)
    except Exception as e:
        logger.error("Impossible to add the template: {0}".format(str(e)))


def _init_xslt():
    """Init the XSLTs. Add XSLTs and the binding with the user template.

    Returns:

    """
    try:
        # Get or create template
        template_version_manager = (
            user_version_manager_api.get_default_version_manager()
        )
        # Get or create XSLTs
        list_xslt = _get_or_create_xslt(LIST_XSL_FILENAME)
        default_detail_xslt = _get_or_create_xslt(DETAIL_XSL_FILENAME)
        list_detail_xslt = [default_detail_xslt]
        # Create binding between template and XSLTs if does not exist
        _bind_template_xslt(
            template_version_manager[0].current,
            list_xslt,
            default_detail_xslt,
            list_detail_xslt,
        )
    except Exception as e:
        print("ERROR : Impossible to init the XSLTs. " + str(e))


def _get_or_create_xslt(filename):
    """Get or create an xslt.

    Args:
        filename: XSLT filename.

    Returns:
        XSLT.

    """
    try:
        return xslt_transformation_api.get_by_name(filename)
    except exceptions.ApiError:
        # Get XSLT.
        list_xslt_path = finders.find(join(XSL_FOLDER_PATH, filename))
        # Read content.
        list_xsl_data = read_file_content(list_xslt_path)
        # Create the XSLT.
        list_xslt = XslTransformation(
            name=filename, filename=filename, content=list_xsl_data
        )
        return xslt_transformation_api.upsert(list_xslt)
    except Exception as e:
        raise Exception(
            "Impossible to add the xslt {0} : {1} ".format(filename, str(e))
        )


def _bind_template_xslt(template_id, list_xslt, default_detail_xslt, list_detail_xslt):
    """Bind the registry template with the XSLTs.

    Args:
        template_id: Registry template id.
        list_xslt: List XSLT.
        default_detail_xslt: Detail XSLT.
        list_detail_xslt:

    Returns:

    """
    from core_main_app.components.template_xsl_rendering import (
        api as template_xsl_rendering_api,
    )

    try:
        template_xsl_rendering_api.get_by_template_id(template_id)
    except exceptions.DoesNotExist:
        template_xsl_rendering_api.add_or_delete(
            template=system_api.get_template_by_id(template_id),
            list_xslt=list_xslt,
            default_detail_xslt=default_detail_xslt,
            list_detail_xslt=list_detail_xslt,
        )
    except Exception as e:
        raise Exception("Impossible to bind the template with XSLTs : " + str(e))


def _init_periodic_tasks():
    """Create periodic tasks for the app and add them to a crontab schedule"""
    schedule, _ = CrontabSchedule.objects.get_or_create(
        minute="*",
    )
    try:
        PeriodicTask.objects.get(name=delete_user_data_structure.__name__)
    except ObjectDoesNotExist:
        PeriodicTask.objects.create(
            crontab=schedule,
            name=delete_user_data_structure.__name__,
            task="core_user_registration_app.tasks.delete_user_data_structure",
        )



"""
User registration app menu
"""

from django.urls import reverse
from menu import Menu, MenuItem


templates_children = (
    MenuItem(
        "User Template List",
        reverse("core-admin:core_user_registration_app_templates"),
        icon="list",
    ),
    MenuItem(
        "Upload New User Template",
        reverse("core-admin:core_user_registration_app_upload_template"),
        icon="upload",
    ),
)

Menu.add_item("admin", MenuItem("USER TEMPLATES", None, children=templates_children))



"""
Template Version Manager model
"""
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from core_main_app.commons import exceptions
from core_main_app.components.template_version_manager.models import (
    TemplateVersionManager,
)


class UserTemplateVersionManager(TemplateVersionManager):
    """Manager of templates versions"""

    # TODO: see if better way to find _cls
    _class_name = "VersionManager.UserTemplateVersionManager"
    is_default = models.BooleanField(default=False)

    @property
    def class_name(self):
        return UserTemplateVersionManager._class_name

    @staticmethod
    def get_by_id(version_manager_id):
        """Return Version Managers by id.

        Args:
            version_manager_id:

        Returns:

        """
        try:
            return UserTemplateVersionManager.objects.get(pk=version_manager_id)
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as e:
            raise exceptions.ModelError(str(e))

    @staticmethod
    def get_global_version_managers(_cls=True):
        """Return all Template Version Managers with user set to None.

        Returns:
            _cls: if True, restricts to TemplateVersionManager _cls
        """
        queryset = UserTemplateVersionManager.objects.filter(user=None).all()
        if _cls:
            queryset = queryset.filter(
                _cls=UserTemplateVersionManager._class_name
            ).all()
        return queryset

    def save_template_version_manager(self):
        """Custom save.

        Returns:
            Saved Instance.

        """
        return super(UserTemplateVersionManager, self).save_version_manager()

    @staticmethod
    def get_default_version_manager():
        """Return default User Version Manager.

        Returns:

        """
        return UserTemplateVersionManager.objects(is_default=True)

    @staticmethod
    def get_active_global_version_manager_by_title(version_manager_title):
        """Return active Template Version Manager by its title with user set to None.

        Args:
            version_manager_title: Version Manager title

        Returns:
            Version Manager instance

        """
        try:
            return UserTemplateVersionManager.objects.get(
                is_disabled=False, title=version_manager_title, user=None
            )
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as e:
            raise exceptions.ModelError(str(e))



""" Access control testing
"""

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_main_app.utils.tests_tools.RequestMock import create_mock_request
from core_user_registration_app.components.user_template_version_manager import (
    api as user_template_vm_api,
)
from tests.components.user_template_version_manager.fixtures.fixtures import (
    UserTemplateVersionManagerAccessControlFixtures,
)

fixture_template_vm = UserTemplateVersionManagerAccessControlFixtures()


# FIXME: missing tests where CAN_ANONYMOUS_ACCESS_PUBLIC_DOCUMENT is True


class TestTemplateVersionManagerInsert(MongoIntegrationBaseTestCase):

    fixture = fixture_template_vm

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_insert_user_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            user_template_vm_api.insert(
                self.fixture.user1_utvm,
                self.fixture.user1_template,
                request=mock_request,
            )

    def test_insert_global_template_as_anonymous_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            user_template_vm_api.insert(
                self.fixture.global_tvm,
                self.fixture.global_template,
                request=mock_request,
            )

    def test_insert_own_template_as_user_saves(self):
        mock_request = create_mock_request(user=self.user1)
        user_template_vm_api.insert(
            self.fixture.user1_utvm, self.fixture.user1_template, request=mock_request
        )

    def test_insert_other_users_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            user_template_vm_api.insert(
                self.fixture.user2_utvm,
                self.fixture.user2_template,
                request=mock_request,
            )

    def test_insert_global_template_as_user_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.user1)
        with self.assertRaises(AccessControlError):
            user_template_vm_api.insert(
                self.fixture.global_tvm,
                self.fixture.global_template,
                request=mock_request,
            )

    def test_insert_own_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        user_template_vm_api.insert(
            self.fixture.user1_utvm, self.fixture.user1_template, request=mock_request
        )

    def test_insert_other_users_template_as_staff_raises_access_control_error(self):
        mock_request = create_mock_request(user=self.staff_user1)
        with self.assertRaises(AccessControlError):
            user_template_vm_api.insert(
                self.fixture.user2_utvm,
                self.fixture.user2_template,
                request=mock_request,
            )

    def test_insert_global_template_as_staff_saves(self):
        mock_request = create_mock_request(user=self.staff_user1)
        user_template_vm_api.insert(
            self.fixture.global_tvm, self.fixture.global_template, request=mock_request
        )

    def test_insert_own_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        user_template_vm_api.insert(
            self.fixture.user1_utvm, self.fixture.user1_template, request=mock_request
        )

    def test_insert_other_users_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        user_template_vm_api.insert(
            self.fixture.user2_utvm, self.fixture.user2_template, request=mock_request
        )

    def test_insert_global_template_as_superuser_saves(self):
        mock_request = create_mock_request(user=self.superuser1)
        user_template_vm_api.insert(
            self.fixture.global_tvm, self.fixture.global_template, request=mock_request
        )


class TestTemplateGetGlobalVersionManagers(MongoIntegrationBaseTestCase):

    fixture = fixture_template_vm

    def setUp(self):
        self.anonymous_user = create_mock_user(user_id=None, is_anonymous=True)
        self.user1 = create_mock_user(user_id="1")
        self.staff_user1 = create_mock_user(user_id="1", is_staff=True)
        self.superuser1 = create_mock_user(user_id="1", is_superuser=True)
        self.fixture.insert_data()

    def test_get_global_version_managers_as_anonymous_raises_acces_control_error(self):
        mock_request = create_mock_request(user=self.anonymous_user)
        with self.assertRaises(AccessControlError):
            user_template_vm_api.get_global_version_managers(request=mock_request)

    def test_get_global_version_managers_as_user_returns_global_tvm(self):
        mock_request = create_mock_request(user=self.user1)
        list_tvm = user_template_vm_api.get_global_version_managers(
            request=mock_request
        )
        for tvm in list_tvm:
            self.assertEqual(tvm.user, None)

    def test_get_global_version_managers_as_staff_returns_global_tvm(self):
        mock_request = create_mock_request(user=self.staff_user1)
        list_tvm = user_template_vm_api.get_global_version_managers(
            request=mock_request
        )
        for tvm in list_tvm:
            self.assertEqual(tvm.user, None)

    def test_get_global_version_managers_as_superuser_returns_global_tvm(self):
        mock_request = create_mock_request(user=self.superuser1)
        list_tvm = user_template_vm_api.get_global_version_managers(
            request=mock_request
        )
        for tvm in list_tvm:
            self.assertEqual(tvm.user, None)



""" Fixtures file for template version manager
"""
from core_main_app.components.template.models import Template
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface
from core_user_registration_app.components.user_template_version_manager.models import (
    UserTemplateVersionManager,
)


class UserTemplateVersionManagerAccessControlFixtures(FixtureInterface):
    """Template Version Manager fixtures"""

    user1_template = None
    user2_template = None
    global_template = None
    user1_utvm = None
    user2_utvm = None
    global_tvm = None
    template_vm_collection = None

    def insert_data(self):
        """Insert a set of Templates and Template Version Managers.

        Returns:

        """
        # Make a connexion with a mock database
        xsd = (
            '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">'
            '<xs:element name="tag"></xs:element></xs:schema>'
        )
        self.user1_utvm = UserTemplateVersionManager(
            title="template 1",
            user="1",
            is_disabled=False,
        )
        self.user1_utvm.save()
        self.user1_template = Template(
            filename="template1.xsd",
            content=xsd,
            hash="hash1",
            user="1",
            version_manager=self.user1_utvm,
            is_current=True,
        )
        self.user1_template.save()
        self.user2_utvm = UserTemplateVersionManager(
            title="template 2",
            user="2",
            is_disabled=False,
        )
        self.user2_utvm.save()
        self.user2_template = Template(
            filename="template2.xsd",
            content=xsd,
            hash="hash2",
            user="2",
            version_manager=self.user2_utvm,
            is_current=True,
        )
        self.user2_template.save()
        self.global_tvm = UserTemplateVersionManager(
            title="global template",
            user=None,
            is_disabled=False,
        )
        self.global_tvm.save()
        self.global_template = Template(
            filename="global_template.xsd",
            content=xsd,
            hash="global hash",
            user=None,
            version_manager=self.global_tvm,
            is_current=True,
        )
        self.global_template.save()

        self.template_vm_collection = [
            self.user1_utvm,
            self.user2_utvm,
            self.global_tvm,
        ]



SECRET_KEY = "fake-key"

INSTALLED_APPS = [
    # Django apps
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    # Extra apps
    "captcha",
    "django_celery_beat",
    # Local apps
    "core_main_app",
    "core_website_app",
    "core_parser_app",
    "core_curate_app",
    "core_user_registration_app",
    "tests",
]

# SERVER URI
SERVER_URI = "http://example.com"

# IN-MEMORY TEST DATABASE
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
        "USER": "",
        "PASSWORD": "",
        "HOST": "",
        "PORT": "",
    },
}

MIDDLEWARE = (
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "tz_detect.middleware.TimezoneMiddleware",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": ["templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "core_main_app.utils.custom_context_processors.domain_context_processor",  # Needed by any curator app
                "django.template.context_processors.i18n",
            ],
        },
    },
]

LOGIN_URL = "/login"
STATIC_URL = "/static/"
ROOT_URLCONF = "tests.urls"

PASSWORD_HASHERS = ("django.contrib.auth.hashers.UnsaltedMD5PasswordHasher",)

DATA_SORTING_FIELDS = ["+title"]

CUSTOM_NAME = "Curator"
ENABLE_SAML2_SSO_AUTH = False
CELERYBEAT_SCHEDULER = "django_celery_beat.schedulers:DatabaseScheduler"



""" System api to access data without access control neither API rules
"""

from core_main_app.components.template.models import Template
from core_main_app.components.version_manager.utils import get_latest_version_name
from core_main_app.utils.xml import is_schema_valid, get_hash
from core_parser_app.components.data_structure.models import DataStructureElement
from core_user_registration_app.components.user_template_version_manager.models import (
    UserTemplateVersionManager,
)


def insert_registry_user_schema(xsd_filename, xsd_content):
    # Check if schema is valid
    is_schema_valid(xsd_content)
    template = Template(
        filename=xsd_filename, hash=get_hash(xsd_content), content=xsd_content
    )
    template.save_template()

    # save the template in database
    try:
        from core_main_app.components.version_manager import api as version_manager_api

        user_version_manager = UserTemplateVersionManager(
            title=xsd_filename, is_default=True
        )
        user_version_manager.save_version_manager()
        # insert the initial template in the version manager
        template.version_manager = user_version_manager
        # set current version
        if len(user_version_manager.versions) == 0:
            template.is_current = True
        # update saved template
        template.display_name = get_latest_version_name(user_version_manager)
        # save template
        template.save()
        # return version manager
        return user_version_manager
    except Exception as e:
        template.delete()
        raise e


def get_all_data_structure_elements():
    """Returns all data structure elements"""
    return DataStructureElement.objects.all()


def get_active_global_version_manager_by_title(version_manager_title):
    """Return all active Version Managers with user set to None.

    Args:
        version_manager_title:
        _cls:

    Returns:

    """
    return UserTemplateVersionManager.get_active_global_version_manager_by_title(
        version_manager_title
    )



""" Tests of the account request API
"""
from unittest.case import TestCase

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist
from mock import Mock, patch

from core_main_app.commons.exceptions import ApiError
from core_user_registration_app.components.account_request_metadata import (
    api as account_request_metadata_api,
)
from core_user_registration_app.components.account_request_metadata.models import (
    AccountRequestMetadata,
)


class TestsAccountRequestMetadataGet(TestCase):
    @patch(
        "core_user_registration_app.components.account_request_metadata.api.get_account_request_metadata_by_id"
    )
    def test_account_request_metadata_get_return_request_object(self, mock_get_by_id):
        # Arrange
        request_id = "1"
        mock_get_by_id.return_value = _create_account_request_metadata()
        # Act
        result = account_request_metadata_api.get_account_request_metadata_by_id(
            request_id
        )
        # Assert
        self.assertIsInstance(result, AccountRequestMetadata)


class TestsAccountRequestMetadataInsert(TestCase):
    def setUp(self):
        self.mock_account_request_metadata = _create_account_request_metadata()

    @patch(
        "core_user_registration_app.components.account_request_metadata.api._get_user_by_username"
    )
    def test_account_request_metadata_insert_raise_ApiError_if_username_already_exist(
        self, mock_get_user_by_username
    ):
        # Arrange
        mock_user = Mock(spec=User)
        mock_user.username = "username"
        mock_get_user_by_username.return_value = mock_user

        # Act # Assert
        with self.assertRaises(ApiError):
            account_request_metadata_api.create_account_request(mock_user)

    @patch(
        "core_user_registration_app.components.account_request_metadata.models.AccountRequestMetadata.save"
    )
    @patch(
        "core_user_registration_app.components.account_request_metadata.api._get_user_by_username"
    )
    def test_account_request_metadata_insert_return_request(
        self, mock_get_user_by_username, mock_save
    ):
        # Arrange
        mock_user = Mock(spec=User)
        mock_user.username = "username"
        mock_get_user_by_username.side_effect = ObjectDoesNotExist()
        mock_save.return_value = self.mock_account_request_metadata

        # Act
        result = account_request_metadata_api.create_account_request(mock_user)

        # Assert
        self.assertIsInstance(result, AccountRequestMetadata)


def _create_account_request_metadata(username="username"):
    """
    Create an AccountRequestMetadata object using default parameters

    Parameters:
        username (str):

    Returns:
        AccountRequestMetadata object
    """
    return AccountRequestMetadata(username=username)



""" Unit Test User Data Structure
"""
from unittest.case import TestCase

from mock import patch

import core_user_registration_app.components.user_data_structure.api as user_data_structure_api
from core_main_app.commons import exceptions
from core_main_app.components.template.models import Template
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)


class TestUserDataStructureGetById(TestCase):
    @patch.object(UserDataStructure, "get_by_id")
    def test_user_data_structure_get_by_id_raises_does_not_exist_error_if_not_found(
        self, mock_get
    ):

        # Arrange
        mock_get.side_effect = exceptions.DoesNotExist("")

        # Act # Assert
        with self.assertRaises(exceptions.DoesNotExist):
            user_data_structure_api.get_by_id(1)

    def test_data_structure_get_by_id_raises_doesnotexist_if_not_found(self):

        # Act # Assert
        with self.assertRaises(exceptions.DoesNotExist):
            user_data_structure_api.get_by_id(1)

    @patch.object(UserDataStructure, "get_by_id")
    def test_user_data_structure_get_by_id_return_data_if_found(self, mock_get):
        # Arrange
        mock_data_structure = UserDataStructure(
            user="1", template=Template(), name="name"
        )
        mock_get.return_value = mock_data_structure
        # Act
        result = user_data_structure_api.get_by_id(1)
        # Assert
        self.assertIsInstance(result, UserDataStructure)


class TestUserDataStructureUpsert(TestCase):
    @patch.object(UserDataStructure, "save_object")
    def test_user_data_structure_upsert_return_data_structure_element(self, mock_save):
        # Arrange
        mock_data_structure = UserDataStructure(
            user="1", template=Template(), name="name"
        )
        mock_save.return_value = mock_data_structure
        mock_user = create_mock_user("1")
        # Act
        result = user_data_structure_api.upsert(mock_data_structure, mock_user)
        # Assert
        self.assertIsInstance(result, UserDataStructure)


class TestUserDataStructureGetAll(TestCase):
    @patch.object(UserDataStructure, "get_all")
    def test_user_data_get_all_return_collection_of_user_data(self, mock_list):
        # Arrange
        mock_data_1 = UserDataStructure(
            user="1", template=_get_template(), name="name_title_1"
        )
        mock_data_2 = UserDataStructure(
            user="1", template=_get_template(), name="name_title_2"
        )
        mock_list.return_value = [mock_data_1, mock_data_2]
        # Act
        result = user_data_structure_api.get_all(
            create_mock_user("1", is_staff=True, is_superuser=True)
        )
        # Assert
        self.assertTrue(all(isinstance(item, UserDataStructure) for item in result))


def _get_template():
    template = Template()
    template.id_field = 1
    xsd = (
        '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">'
        '<xs:element name="tag"></xs:element></xs:schema>'
    )
    template.content = xsd
    return template



""" AccountRequestMetadata api
"""

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist

import core_main_app.utils.notifications.mail as send_mail_api
from core_main_app.access_control.api import (
    has_perm_administration,
)
from core_main_app.access_control.decorators import access_control
from core_main_app.commons.exceptions import ApiError
from core_user_registration_app.components.account_request_metadata.models import (
    AccountRequestMetadata,
)
from core_website_app.settings import SERVER_URI


@access_control(has_perm_administration)
def get_all(user):
    """List of opened account requests

    Returns:

        List of all requests
    """
    return AccountRequestMetadata.get_all()


def _get_user_by_username(username):
    """Returns a user given its username

    Args:

        username: Given username

    Returns:

        User
    """
    return User.objects.get(username=username)


def get_account_request_metadata_by_id(account_id):
    """Returns an account Request metadata given its id


    Args:
        account_id:

    Returns:
        AccountRequestMetadata

    """

    return AccountRequestMetadata.get_by_id(account_id)


def _get_user_by_id(user_id):
    """Returns a user given its primary key

    Args:

        user_id: Given user id

    Returns:

        User
    """
    return User.objects.get(pk=user_id)


def insert_metadata(user, account_id, metadata):
    """Insert metadata into RequestAccountMetadata given its id

    Args:
        user
        account_id
        metadata


    """
    # NOTE: used queryset.update to change metadata as assignment would not save change in database
    AccountRequestMetadata.objects.filter(pk=account_id).update(metadata=metadata)


def create_account_request(user):
    """Create a new request with the metadata

    Args:
        user: Django User

    Returns: New Account request

    """

    try:
        # check if a user with the same username exists
        _get_user_by_username(user.username)
        raise ApiError("A user with the same username already exists.")
    except ObjectDoesNotExist:
        user.save()

        # Create the account request and save it
        account_request = AccountRequestMetadata(
            username=user.username,
            first_name=user.first_name,
            last_name=user.last_name,
            email=user.email,
        )

        context = {"URI": SERVER_URI}
        template_path = "core_website_app/admin/email/request_account_for_admin.html"
        send_mail_api.send_mail_to_administrators(
            subject="New Account Request",
            path_to_template=template_path,
            context=context,
        )
        account_request.save()
        return account_request



"""User Registration tasks """

import logging
from datetime import timedelta

from celery import shared_task
from django.utils import timezone

from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)
from core_user_registration_app.settings import (
    USER_DATA_STRUCTURE_HOURS_THRESHOLD,
)
from core_user_registration_app.system import api as system_api

logger = logging.getLogger(__name__)


@shared_task
def delete_user_data_structure():
    """DELETES every DELETE_USER_DATA_STRUCTURE_RATE the UserDataStructure in the DataStructure collection"""
    logger.info("Checking Old UserDataStructures")
    try:
        user_data_structure_perm = UserDataStructure.get_permission()
        for data_structure_element in system_api.get_all_data_structure_elements():
            if (
                data_structure_element.data_structure
                and data_structure_element.data_structure.get_object_permission()
                == user_data_structure_perm
            ):
                if data_structure_element.creation_date < timezone.now() - timedelta(
                    hours=USER_DATA_STRUCTURE_HOURS_THRESHOLD
                ):
                    data_structure_element.delete()
        logger.info("FINISH checking DataStructures.")

    except Exception as e:
        logger.error(f"ERROR : Error while deleting data structures: {str(e)}")



""" Url router for the administration site
"""
from django.contrib import admin
from django.urls import re_path

from core_main_app.admin import core_admin_site
from core_user_registration_app.components.account_request_metadata.models import (
    AccountRequestMetadata,
)
from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)
from core_user_registration_app.components.user_metadata.models import UserMetadata
from core_user_registration_app.components.user_template_version_manager.models import (
    UserTemplateVersionManager,
)
from core_user_registration_app.views.admin import (
    views as admin_views,
    ajax as admin_ajax,
)

admin_urls = [
    re_path(
        r"^user-template/version/(?P<pk>\w+)/current/$",
        admin_ajax.CurrentTemplateVersion.as_view(),
        name="core_user_registration_app_template_version_current",
    ),
    re_path(
        r"^user-template$",
        admin_views.manage_user_templates,
        name="core_user_registration_app_templates",
    ),
    re_path(
        r"^user-template/versions/(?P<version_manager_id>\w+)",
        admin_views.manage_user_template_versions,
        name="core_user_registration_app_manage_template_versions",
    ),
    re_path(
        r"user-template/upload",
        admin_views.upload_template,
        name="core_user_registration_app_upload_template",
    ),
    # Overrides user-requests
    re_path(
        r"^user-registration-requests$",
        admin_views.user_requests,
        name="core_website_app_user_requests",
    ),
    re_path(
        r"^metadata",
        admin_views.ViewMetaData.as_view(
            administration=True,
            template="core_main_registry_app/admin/data/view_data.html",
        ),
        name="core_user_registration_app_metadata_detail",
    ),
    re_path(
        r"^template/version/current",
        admin_ajax.set_current_template_version_from_version_manager,
        name="core_user_registration_app_set_current_template_version",
    ),
]

admin.site.register(UserDataStructure)
admin.site.register(AccountRequestMetadata)
admin.site.register(UserMetadata)
admin.site.register(UserTemplateVersionManager)
urls = core_admin_site.get_urls()
core_admin_site.get_urls = lambda: admin_urls + urls



from django.contrib import messages
from django.contrib.auth.hashers import make_password
from django.contrib.auth.models import User
from django.core.exceptions import ValidationError
from django.shortcuts import redirect
from django.template.loader import get_template
from django.urls import reverse
from django.views import View

from core_curate_app.views.user.views import generate_form, render_form
from core_main_app.commons.exceptions import ApiError
from core_main_app.utils.rendering import render
from core_user_registration_app.components.account_request_metadata import (
    api as account_request_metadata_api,
)
from core_user_registration_app.components.user_data_structure import (
    api as user_data_structure_api,
)
from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)
from core_user_registration_app.components.user_template_version_manager import (
    api as user_version_manager_api,
)
from core_website_app.views.user.forms import RequestAccountForm


def request_new_account(request):
    assets = {
        "js": [
            {"path": "core_website_app/user/js/user_account_req.js", "is_raw": False}
        ],
        "css": ["core_website_app/user/css/list.css"],
    }

    if request.method == "POST":
        request_form = RequestAccountForm(request.POST)
        if request_form.is_valid():
            # Call the API
            try:
                request_form_data = request_form.cleaned_data
                version_manager = user_version_manager_api.get_default_version_manager()

                template = version_manager[0].current_version
                name = request.POST["username"]

                user = User(
                    username=request_form_data.get("username"),
                    first_name=request_form_data.get("firstname"),
                    last_name=request_form_data.get("lastname"),
                    password=make_password(request_form_data.get("password1")),
                    email=request_form_data.get("email"),
                    is_active=False,
                )

                user_data_structure = UserDataStructure(
                    user=name, template=template, name=name
                )

                user_data_structure_api.upsert(user_data_structure, user)
                objectid = user_data_structure.id

                account_request = account_request_metadata_api.create_account_request(
                    user
                )
                messages.add_message(
                    request,
                    messages.INFO,
                    "User Account Request sent to the administrator.",
                )

                url = reverse(
                    "core_user_registration_app_account_metadata",
                    args=(objectid, account_request.id),
                )
                return redirect(url)

            except ApiError as e:
                error_message = str(e)

                error_template = get_template(
                    "core_website_app/user/request_error.html"
                )
                error_box = error_template.render({"error_message": error_message})

                return render(
                    request,
                    "core_website_app/user/request_new_account.html",
                    assets=assets,
                    context={"request_form": request_form, "action_result": error_box},
                )
            except ValidationError as e:
                error_message = "The following error(s) occurred during " "validation:"
                error_items = [str(error) for error in e.messages]

                error_template = get_template(
                    "core_website_app/user/request_error.html"
                )
                error_box = error_template.render(
                    {"error_message": error_message, "error_items": error_items}
                )

                return render(
                    request,
                    "core_website_app/user/request_new_account.html",
                    assets=assets,
                    context={"request_form": request_form, "action_result": error_box},
                )

    else:
        request_form = RequestAccountForm()

    context = {"request_form": request_form}

    return render(
        request,
        "core_user_registration_app/user/request_new_account.html",
        assets=assets,
        context=context,
    )


class AccountCreationView(View):
    def __init__(self):
        super(AccountCreationView, self).__init__()
        self.assets = {
            "js": [
                {
                    "path": "core_user_registration_app/user/js/save_data.js",
                    "is_raw": False,
                },
                {
                    "path": "core_user_registration_app/user/js/save_data.raw.js",
                    "is_raw": False,
                },
                {
                    "path": "core_parser_app/js/autosave.js",
                    "is_raw": False,
                },
                {
                    "path": "core_user_registration_app/user/js/autosave.raw.js",
                    "is_raw": True,
                },
                {
                    "path": "core_user_registration_app/user/js/report_data.js",
                    "is_raw": False,
                },
                {
                    "path": "core_user_registration_app/user/js/report_data.raw.js",
                    "is_raw": True,
                },
            ],
            "css": ["core_curate_app/user/css/xsd_form.css"],
        }

    def get(self, request, objectid, accountid):
        context = self.build_context(request, objectid, accountid)
        return render(
            request,
            "core_user_registration_app/user/account_creation.html",
            assets=self.assets,
            context=context,
        )

    def build_context(self, request, objectid, account_id):
        """Builds the context.

        Args:
            request:
            objectid:

        Returns:

        """
        #
        user_data_structure = user_data_structure_api.get_by_id(objectid)
        xsd_string = user_data_structure.template.content
        xml_string = user_data_structure.form_string
        root_element = generate_form(
            xsd_string, xml_string, user_data_structure, request
        )
        user_data_structure_api.update_data_structure_root(
            user_data_structure, root_element, request.user
        )
        xsd_form = render_form(request, root_element)

        return {
            "edit": True if user_data_structure.data is not None else False,
            "xsd_form": xsd_form,
            "data_structure": user_data_structure,
            "account_id": account_id,
        }



""" Metadata model
"""
from django.core.exceptions import ObjectDoesNotExist
from django.db import models

from core_main_app.commons import exceptions
from core_main_app.components.abstract_data.models import AbstractData
from core_main_app.components.template.models import Template
from core_main_app.components.workspace.models import Workspace


class UserMetadata(AbstractData):
    """Metadata object"""

    template = models.ForeignKey(Template, blank=False, on_delete=models.CASCADE)
    user_id = models.CharField(max_length=200)
    workspace = models.ForeignKey(
        Workspace, blank=True, on_delete=models.SET_NULL, null=True
    )

    class Meta:
        verbose_name = "User metadata"
        verbose_name_plural = "User metadata"
        indexes = [
            models.Index(
                fields=["title", "last_modification_date", "template", "user_id"]
            ),
        ]

    def get_dict_content(self):
        """Return dict_content

        Returns:

        """
        return self.dict_content

    @staticmethod
    def get_all(order_by_field):
        """Get all data.

        Args:
            order_by_field: Order by field.

        Returns:

        """
        return UserMetadata.objects.all().order_by(
            *[field.replace("+", "") for field in order_by_field]
        )

    @staticmethod
    def get_by_id(data_id):
        """Return the object with the given id.

        Args:
            data_id:

        Returns:
            Data (obj): Data object with the given id

        """
        try:
            return UserMetadata.objects.get(pk=str(data_id))
        except ObjectDoesNotExist as e:
            raise exceptions.DoesNotExist(str(e))
        except Exception as ex:
            raise exceptions.ModelError(str(ex))

    def __str__(self):
        """User metadata object as string

        Returns:

        """
        return self.title



""" ACL Test Data Structure
"""

from django.contrib.auth.models import AnonymousUser

import core_user_registration_app.components.user_data_structure.api as user_data_structure_api
from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.utils.integration_tests.integration_base_test_case import (
    MongoIntegrationBaseTestCase,
)
from core_main_app.utils.tests_tools.MockUser import create_mock_user
from core_parser_app.components.data_structure.models import DataStructureElement
from core_user_registration_app.components.user_data_structure.models import (
    UserDataStructure,
)
from tests.components.user_data_structure.fixtures.fixtures import (
    UserDataStructureFixtures,
)

fixture_data_structure = UserDataStructureFixtures()


class TestUserDataStructureGetAll(MongoIntegrationBaseTestCase):
    fixture = fixture_data_structure

    def test_get_all_as_superuser_returns_all_data_structure(self):
        mock_user = create_mock_user("1", is_staff=True, is_superuser=True)
        result = user_data_structure_api.get_all(mock_user)
        self.assertTrue(all(isinstance(item, UserDataStructure) for item in result))

    def test_get_all_as_user_raises_error(self):
        mock_user = create_mock_user("1")
        with self.assertRaises(AccessControlError):
            user_data_structure_api.get_all(mock_user)

    def test_get_all_as_anonymous_user_raises_error(self):
        with self.assertRaises(AccessControlError):
            user_data_structure_api.get_all(AnonymousUser())


class TestUserDataStructureDelete(MongoIntegrationBaseTestCase):
    fixture = fixture_data_structure

    def test_delete_others_data_structure_as_superuser_deletes_data_structure(self):
        data_structure = self.fixture.data_structure_3
        mock_user = create_mock_user(
            self.fixture.data_structure_1.user, is_staff=True, is_superuser=True
        )
        user_data_structure_api.delete(data_structure, mock_user)

    def test_delete_own_data_structure_deletes_data_structure(self):
        data_structure = self.fixture.data_structure_1
        mock_user = create_mock_user(self.fixture.data_structure_1.user)
        user_data_structure_api.delete(data_structure, mock_user)


class TestUserDataStructureUpdateDataStructureRoot(MongoIntegrationBaseTestCase):
    fixture = fixture_data_structure

    def test_update_others_data_structure_root_as_superuser_updates_data_structure(
        self,
    ):
        data_structure = self.fixture.data_structure_1
        new_data_structure_element_root = DataStructureElement()
        new_data_structure_element_root.save()
        mock_user = create_mock_user(
            self.fixture.data_structure_2.user, is_staff=True, is_superuser=True
        )
        user_data_structure_api.update_data_structure_root(
            data_structure, new_data_structure_element_root, mock_user
        )
        result = user_data_structure_api.get_by_id(data_structure.id)
        self.assertTrue(isinstance(result, UserDataStructure))
        self.assertTrue(
            result.data_structure_element_root, new_data_structure_element_root
        )

    def test_update_own_data_structure_root_updates_data_structure(self):
        data_structure = self.fixture.data_structure_1
        new_data_structure_element_root = DataStructureElement()
        new_data_structure_element_root.save()
        mock_user = create_mock_user(self.fixture.data_structure_1.user)
        user_data_structure_api.update_data_structure_root(
            data_structure, new_data_structure_element_root, mock_user
        )
        result = user_data_structure_api.get_by_id(data_structure.id)
        self.assertTrue(isinstance(result, UserDataStructure))
        self.assertTrue(
            result.data_structure_element_root, new_data_structure_element_root
        )


class TestUserDataStructureCreateOrUpdate(MongoIntegrationBaseTestCase):
    fixture = fixture_data_structure

    def test_upsert_own_data_structure_updates_data_structure(self):
        data_structure = self.fixture.data_structure_1
        data_structure.name = "new_name_1"
        mock_user = create_mock_user(self.fixture.data_structure_1.user)
        user_data_structure_api.upsert(data_structure, mock_user)
        result = user_data_structure_api.get_by_id(data_structure.id)
        self.assertTrue(isinstance(result, UserDataStructure))
        self.assertTrue(data_structure.name, result.name)


class TestUserDataStructureGetByDataId(MongoIntegrationBaseTestCase):
    fixture = fixture_data_structure

    def test_get_by_data_id_as_superuser_returns_data_structure(self):

        mock_user = create_mock_user(
            self.fixture.data_structure_3.user, is_staff=True, is_superuser=True
        )
        data_structure = user_data_structure_api.get_by_data_id(
            self.fixture.data.id, mock_user
        )
        self.assertTrue(isinstance(data_structure, UserDataStructure))
        self.assertEquals(self.fixture.data.id, data_structure.data.id)

    def test_get_by_data_id_as_owner_returns_data_structure(self):
        mock_user = create_mock_user(self.fixture.data_structure_1.user)
        data_structure = user_data_structure_api.get_by_data_id(
            self.fixture.data.id, mock_user
        )
        self.assertTrue(isinstance(data_structure, UserDataStructure))
        self.assertEquals(self.fixture.data.id, data_structure.data.id)

    def test_get_by_data_id_as_user_non_owner_raises_error(self):
        mock_user = create_mock_user(self.fixture.data_structure_3.user)
        with self.assertRaises(AccessControlError):
            user_data_structure_api.get_by_data_id(self.fixture.data.id, mock_user)

    def test_get_by_data_id_as_anonymous_user_raises_error(self):
        with self.assertRaises(AccessControlError):
            user_data_structure_api.get_by_data_id(
                self.fixture.data.id, AnonymousUser()
            )


class TestDataStructureChangeOwner(MongoIntegrationBaseTestCase):

    fixture = fixture_data_structure

    def test_change_owner_from_owner_to_owner_ok(self):
        mock_owner = create_mock_user(self.fixture.data_structure_1.user)
        user_data_structure_api.change_owner(
            document=fixture_data_structure.data_structure_1,
            new_user=mock_owner,
            user=mock_owner,
        )

    def test_change_owner_from_owner_to_user_ok(self):
        mock_owner = create_mock_user(self.fixture.data_structure_1.user)
        mock_user = create_mock_user("2")
        user_data_structure_api.change_owner(
            document=fixture_data_structure.data_structure_1,
            new_user=mock_user,
            user=mock_owner,
        )

    def test_change_owner_from_user_to_user_raises_exception(self):
        mock_owner = create_mock_user("0")
        mock_user = create_mock_user("2")
        with self.assertRaises(AccessControlError):
            user_data_structure_api.change_owner(
                document=fixture_data_structure.data_structure_1,
                new_user=mock_user,
                user=mock_owner,
            )

    def test_change_owner_from_anonymous_to_user_raises_exception(self):

        mock_user = create_mock_user("2")
        with self.assertRaises(AccessControlError):
            user_data_structure_api.change_owner(
                document=fixture_data_structure.data_structure_1,
                new_user=mock_user,
                user=AnonymousUser(),
            )

    def test_change_owner_as_superuser_ok(self):
        mock_user = create_mock_user("2", is_superuser=True)
        user_data_structure_api.change_owner(
            document=fixture_data_structure.data_structure_1,
            new_user=mock_user,
            user=mock_user,
        )



""" Fixtures files for UserMetadata
"""
from core_main_app.components.template.models import Template
from core_main_app.components.workspace import api as workspace_api
from core_main_app.components.workspace.models import Workspace
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface
from core_user_registration_app.components.user_metadata.models import UserMetadata


class DataFixtures(FixtureInterface):
    """UserMetadata fixtures"""

    data_1 = None
    data_2 = None
    data_3 = None
    template = None
    data_collection = None

    def insert_data(self):
        """Insert a set of UserMetadata.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_template()
        self.generate_data_collection()

    def generate_data_collection(self):
        """Generate a UserMetadata collection.

        Returns:

        """
        # NOTE: no xml_content to avoid using unsupported GridFS mock
        self.data_1 = UserMetadata(
            template=self.template, user_id="1", dict_content=None, title="title"
        )
        self.data_1.save()
        self.data_2 = UserMetadata(
            template=self.template, user_id="2", dict_content=None, title="title2"
        )
        self.data_2.save()
        self.data_3 = UserMetadata(
            template=self.template, user_id="1", dict_content=None, title="title3"
        )
        self.data_3.save()
        self.data_collection = [self.data_1, self.data_2, self.data_3]

    def generate_template(self):
        """Generate an unique Template.

        Returns:

        """
        self.template = Template()
        xsd = (
            '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">'
            '<xs:element name="tag"></xs:element></xs:schema>'
        )
        self.template.content = xsd
        self.template.hash = ""
        self.template.filename = "filename"
        self.template.save()


class AccessControlDataFixture(FixtureInterface):
    """Access Control UserMetadata fixture"""

    USER_1_NO_WORKSPACE = 0
    USER_2_NO_WORKSPACE = 1
    USER_1_WORKSPACE_1 = 2
    USER_2_WORKSPACE_2 = 3

    template = None
    workspace_1 = None
    workspace_2 = None
    data_collection = None
    data_1 = None
    data_2 = None
    data_3 = None
    data_4 = None
    data_5 = None

    def insert_data(self):
        """Insert a set of UserMetadata.

        Returns:

        """
        # Make a connexion with a mock database
        self.generate_template()
        self.generate_workspace()
        self.generate_data_collection()

    def generate_data_collection(self):
        """Generate a UserMetadata collection.

        Returns:

        """

        content = {"root": {"element": "value2"}}

        self.data_1 = UserMetadata(template=self.template, title="Data 1", user_id="1")
        self.data_1.save()
        self.data_2 = UserMetadata(template=self.template, title="Data 2", user_id="2")
        self.data_2.save()
        self.data_3 = UserMetadata(
            template=self.template,
            title="Data 3",
            user_id="1",
            workspace=self.workspace_1,
            dict_content=content,
        )
        self.data_3.save()
        self.data_4 = UserMetadata(
            template=self.template,
            title="DataDoubleTitle",
            user_id="2",
            workspace=self.workspace_2,
        )
        self.data_4.save()
        self.data_5 = UserMetadata(
            template=self.template,
            title="DataDoubleTitle",
            user_id="1",
            workspace=self.workspace_1,
        )
        self.data_5.save()

        self.data_collection = [
            self.data_1,
            self.data_2,
            self.data_3,
            self.data_4,
            self.data_5,
        ]

    def generate_template(self):
        """Generate an unique Template.

        Returns:

        """
        self.template = Template()
        xsd = (
            '<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">'
            '<xs:element name="tag"></xs:element></xs:schema>'
        )
        self.template.content = xsd
        self.template.hash = ""
        self.template.filename = "filename"
        self.template.save()

    def generate_workspace(self):
        """Generate the workspaces.

        Returns:

        """
        self.workspace_1 = Workspace(
            title="Workspace 1", owner="1", read_perm_id="1", write_perm_id="1"
        )
        self.workspace_1.save()
        self.workspace_2 = Workspace(
            title="Workspace 2", owner="2", read_perm_id="2", write_perm_id="2"
        )
        self.workspace_2.save()



"""
Django settings for core_user_registration_app project.

Generated by 'django-admin startproject' using Django 1.11.

For more information on this file, see
https://docs.djangoproject.com/en/1.11/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/1.11/ref/settings/
"""

from os.path import join

from django.conf import settings

if not settings.configured:
    settings.configure()

REGISTRY_XSD_FILENAME = getattr(settings, "REGISTRY_XSD_FILENAME", "")
""" str: Registry xsd filename used for the initialisation.
"""

XSL_FOLDER_PATH = getattr(
    settings, "XSL_FOLDER_PATH", "core_explore_keyword_registry_app/xsl"
)
""" str: Xsl folder path used for the initialisation.
"""

LIST_XSL_FILENAME = getattr(settings, "LIST_XSL_FILENAME", "registry-list.xsl")
"""" str : List xsl filename used for the initialisation.
"""

DETAIL_XSL_FILENAME = getattr(settings, "DETAIL_XSL_FILENAME", "registry-detail.xsl")
"""  str : Detail xsl filename used for the initialisation.
"""

REGISTRY_XSD_USER_FILEPATH = getattr(
    settings,
    "REGISTRY_USER_XSD_FILEPATH",
    join("core_user_registration_app", "user", "xsd", "user.xsd"),
)
""" str: Registry xsd path used for the initialisation.
"""

REGISTRY_XSD_USER_FILENAME = getattr(settings, "REGISTRY_XSD_USER_FILENAME", "user.xsd")
""" str: Registry xsd filename used for the initialisation.
"""


DELETE_USER_DATA_STRUCTURE_RATE = 10
""" int: Task rate in seconds.
"""

USER_DATA_STRUCTURE_HOURS_THRESHOLD = 1
""" int: Number of hours before deleting a User Data Structure
"""



"""AJAX views
"""

import json

from django.contrib import messages
from django.http.response import HttpResponseBadRequest, HttpResponse

from core_main_app.access_control.exceptions import AccessControlError
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.components.workspace import api as workspace_api
from core_main_app.utils.xml import validate_xml_data
from core_parser_app.components.data_structure_element import (
    api as data_structure_element_api,
)
from core_parser_app.tools.parser.renderer.xml import XmlRenderer
from core_user_registration_app.components.account_request_metadata import (
    api as account_request_metadata_api,
)
from core_user_registration_app.components.user_data_structure import (
    api as user_data_structure_api,
)
from core_user_registration_app.components.user_metadata import api as data_api
from core_user_registration_app.components.user_metadata.models import UserMetadata
from xml_utils.xsd_tree.xsd_tree import XSDTree


def save_data(request):
    response_dict = {}
    try:

        # get user data structure
        user_data_structure_id = request.POST["id"]
        account_request_metadata_id = request.POST["metadata"]
        user_data_structure = user_data_structure_api.get_by_id(user_data_structure_id)

        # generate xml data
        xml_data = render_xml(user_data_structure.data_structure_element_root, request)

        # update user data structure data
        user_data_structure.form_string = xml_data

        # save data structure
        user_data_structure_api.upsert(user_data_structure, request.user)

        # build trees
        xsd_tree = XSDTree.build_tree(user_data_structure.template.content)
        xml_tree = XSDTree.build_tree(xml_data)

        # validate XML document
        errors = validate_xml_data(xsd_tree, xml_tree, request=request)
        if errors is not None:
            response_dict["errors"] = errors

        if errors is None:
            # generate the XML
            xml_data = render_xml(
                user_data_structure.data_structure_element_root, request
            )

            # create new data
            data = UserMetadata()
            data.title = user_data_structure.name
            data.template = user_data_structure.template
            data.user_id = str(request.user.id)
            data.id = user_data_structure.pk
            data.workspace = workspace_api.get_global_workspace()

            # set content
            data.xml_content = xml_data
            # save data
            data_api.upsert(data, request.user)
            # insert Metadata reference in account request
            account_request_metadata_api.insert_metadata(
                request.user, account_request_metadata_id, data
            )
            user_data_structure_api.delete(user_data_structure, request.user)
            messages.add_message(
                request, messages.SUCCESS, "User metadata saved with success."
            )
    except Exception as e:
        message = str(e).replace('"', "'")
        return HttpResponseBadRequest(message, content_type="application/javascript")

    return HttpResponse(
        json.dumps({"data_id": str(data.id)}), content_type="application/javascript"
    )


def render_xml(root_element, request):
    """Render the XML.

    Args:
        root_element:

    Returns:

    """
    # build XML renderer
    xml_renderer = XmlRenderer(root_element, request)

    # generate xml data
    xml_data = xml_renderer.render()

    return xml_data


def data_structure_element_value(request):
    """Endpoint for data structure element value

    Args:
        request:

    Returns:

    """
    if request.method == "GET":
        return get_data_structure_element_value(request)
    elif request.method == "POST":
        return save_data_structure_element_value(request)


def get_data_structure_element_value(request):
    """Gets the value of a data structure element

    Args:
        request:

    Returns:

    """
    if "id" not in request.GET:
        return HttpResponseBadRequest()

    try:
        element = data_structure_element_api.get_by_id(request.GET["id"], request)
        element_value = element.value

        if element.tag == "module":
            element_value = {
                "data": element.options["data"],
                "attributes": element.options["attributes"],
            }

        return HttpResponse(
            json.dumps({"value": element_value}), content_type="application/json"
        )
    except (AccessControlError, DoesNotExist) as exc:
        return HttpResponseBadRequest(json.dumps({"message": str(exc)}))


def save_data_structure_element_value(request):
    """Saves the value of a data structure element

    Args:
        request:

    Returns:

    """
    if "id" not in request.POST or "value" not in request.POST:
        return HttpResponseBadRequest(
            "Error when trying to data structure element: id or value is missing."
        )

    try:
        input_element = data_structure_element_api.get_by_id(
            request.POST["id"], request
        )

        input_previous_value = input_element.value
        input_element.value = request.POST["value"]
        data_structure_element_api.upsert(input_element, request)

        return HttpResponse(
            json.dumps({"replaced": input_previous_value}),
            content_type="application/json",
        )
    except (AccessControlError, DoesNotExist) as exc:
        return HttpResponseBadRequest(json.dumps({"message": str(exc)}))



"""core_user_registration_app URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.11/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.conf.urls import url, include
    2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))
"""
from django.conf.urls import include
from django.urls import re_path

import core_user_registration_app.views.admin.ajax as admin_ajax
import core_user_registration_app.views.user.ajax as user_ajax
import core_user_registration_app.views.user.views as registration_views
from core_main_app.rest.template import views as template_views

urlpatterns = [
    re_path("captcha/", include("captcha.urls")),
    re_path(
        r"^account-request",
        registration_views.request_new_account,
        name="core_user_registration_app_account_request",
    ),
    re_path(
        r"^request-metadata/(?P<objectid>\w+)/(?P<accountid>\w+)$",
        registration_views.AccountCreationView.as_view(),
        name="core_user_registration_app_account_metadata",
    ),
    re_path(
        r"^request-metadata/data-structure-element/value$",
        user_ajax.data_structure_element_value,
        name="core_user_registration_app_data_structure_element_value",
    ),
    re_path(
        r"^template/version/current",
        admin_ajax.set_current_template_version_from_version_manager,
        name="core_user_registration_app_set_current_template_version",
    ),
    re_path(
        r"^save-data$", user_ajax.save_data, name="core_user_registration_app_save_data"
    ),
    re_path(
        r"^template/(?P<pk>\w+)/download/$",
        template_views.TemplateDownload.as_view(),
        name="core_main_app_rest_template_download",
    ),
]



""" Metadata API
"""
import datetime

import pytz

from core_main_app.commons import exceptions as exceptions
from core_main_app.utils.xml import validate_xml_data
from core_user_registration_app.components.user_metadata.models import UserMetadata
from xml_utils.xsd_tree.xsd_tree import XSDTree


def get_by_id(data_id, user):
    """Return data object with the given id.

    Parameters:
        data_id:
        user:

    Returns: data object
    """
    return UserMetadata.get_by_id(data_id)


def upsert(data, request):
    """Save or update the data.

    Args:
        data:
        user:

    Returns:

    """
    if data.xml_content is None:
        raise exceptions.ApiError("Unable to save data: xml_content field is not set.")

    data.last_modification_date = datetime.datetime.now(pytz.utc)
    check_xml_file_is_valid(data, request=request)
    return data.convert_and_save()


def check_xml_file_is_valid(data, request):
    """Check if xml data is valid against a given schema.

    Args:
        data:
        request:

    Returns:

    """
    template = data.template

    try:
        xml_tree = XSDTree.build_tree(data.xml_content)
    except Exception as e:
        raise exceptions.XMLError(str(e))

    try:
        xsd_tree = XSDTree.build_tree(template.content)
    except Exception as e:
        raise exceptions.XSDError(str(e))
    error = validate_xml_data(xsd_tree, xml_tree, request=request)
    if error is not None:
        raise exceptions.XMLError(error)
    else:
        return True



""" Fixtures files for User
"""
from django.contrib.auth.models import User

import core_main_app.permissions.rights as rights
from core_main_app.permissions import api as permissions_api
from core_main_app.utils.integration_tests.fixture_interface import FixtureInterface


class UserFixtures(FixtureInterface):
    """User Fixture"""

    def insert_data(self):
        pass

    @staticmethod
    def create_user(username="username", password=None, email=None, is_staff=False):
        user = User.objects.create_user(
            username=username, password=password, email=email, is_staff=is_staff
        )
        user.save()
        return user

    @staticmethod
    def create_super_user(username="username", password="pass", email="email"):
        user = User.objects.create_superuser(username, password, email)
        user.save()
        return user

    @staticmethod
    def add_publish_perm(user):
        publish_perm = permissions_api.get_by_codename(rights.publish_data)
        user.user_permissions.add(publish_perm)
        user.save()
